__all__ = [
	'__author__',
	'__author_email__',
	'__copyright__',
	'__license__',
	'__summary__',
	'__title__',
	'__url__',
	'__version__',
	'__version_info__',
]

try:
	from importlib.metadata import metadata
except ImportError:
	from importlib_metadata import metadata

meta = metadata('google-music')

__title__ = meta['Name']
__summary__ = meta['Summary']
__url__ = meta['Home-page']

__version__ = meta['Version']
__version_info__ = tuple(int(i) for i in __version__.split('.') if i.isdigit())

__author__ = meta['Author']
__author_email__ = meta['Author-email']

__license__ = meta['License']
__copyright__ = f'2018-2020 {__author__} <{__author_email__}>'



__all__ = ['create_mac_string', 'get_ple_prev_next']


def create_mac_string(mac_int, *, delimiter=':'):
	mac = hex(mac_int)[2:].upper()
	pad = max(12 - len(mac), 0)
	mac += '0' * pad

	return delimiter.join(
		mac[x : x + 2]
		for x in range(0, 12, 2)
	)


def get_ple_prev_next(
	playlist_songs,
	*,
	after=None,
	before=None,
	index=None,
	position=None
):
	pl_songs_len = len(playlist_songs)

	if (
		(after or before)
		and position
	):
		raise ValueError(
			"Must provide one or both of 'after'/'before' or one of 'index'/'position'."
		)

	if (
		index is not None
		and index not in range(0, pl_songs_len + 1)
	):
		raise ValueError(f"'index' must be between 0 and {pl_songs_len}.")

	if (
		position is not None
		and position not in range(1, pl_songs_len + 2)
	):
		raise ValueError(f"'position' must be between 1 and {pl_songs_len + 1}.")

	prev = after or {}
	next_ = before or {}

	if prev or next_:
		if prev:
			index = playlist_songs.index(prev)
			if not next_ and prev != playlist_songs[-1]:
				index += 1
				next_ = playlist_songs[index]

		if next_:
			index = playlist_songs.index(next_)
			if not prev and next_ != playlist_songs[0]:
				prev = playlist_songs[index]
	else:
		if position is not None:
			index = position - 1
		elif (
			index is None
			or index == pl_songs_len
		):
			index = pl_songs_len

		if index != 0:
			prev = playlist_songs[index - 1]

		if index != pl_songs_len:
			next_ = playlist_songs[index]

	return prev, next_



import time

from tenacity import retry, stop_after_attempt, wait_exponential

from ..sessions import GoogleMusicSession
from ..token_handlers import FileTokenHandler

try:
	from httpx import HTTPError as RequestError
except ImportError:
	from httpx import RequestError


# TODO: Configurable token updater/saver/loader.
class GoogleMusicClient:
	def __init__(
		self,
		username,
		*,
		session=None,
		token=None,
		token_handler=FileTokenHandler,
		token_handler_kwargs=None
	):
		self._username = username or ''

		if token_handler_kwargs is None:
			token_handler_kwargs = {}

		self._token_handler = token_handler(
			username=self._username,
			client=self.client,
			**token_handler_kwargs
		)

		self._session = (
			session
			or GoogleMusicSession(
				self.client_id,
				self.client_secret,
				self.oauth_scope,
				token=token
			)
		)

	@property
	def is_authenticated(self):
		"""The authentication status of the client instance."""

		return self._session.authorized

	@property
	def token(self):
		return self._session.token

	@token.setter
	def token(self, token):
		self._session.token = token

	@property
	def username(self):
		"""The username associated with the client instance.

		This is used to store OAuth credentials for different accounts separately.
		"""

		return self._username

	@retry(
		reraise=True,
		stop=stop_after_attempt(5),
		wait=wait_exponential(multiplier=1, max=10),
	)
	def _call(self, call_cls, *args, **kwargs):
		call = call_cls(*args, **kwargs)

		# Override default hl/tier params from google-music-proto for Mobileclient.
		params = {**call.params, **self._session.params}

		response = self._session.request(
			call.method,
			call.url,
			headers=call.headers,
			data=call.body,
			params=params,
			allow_redirects=call.follow_redirects
		)

		self._token_handler.dump(self.token)

		try:
			response.raise_for_status()
		except RequestError:
			raise

		return call.parse_response(response.headers, response.content)

	def login(self):
		"""Log in to Google Music.

		Parameters:
			username (str, Optional):
				Your Google Music username.
				Used to store OAuth tokens for multiple accounts separately.
			token (dict, Optional):
				An OAuth token compatible with ``oauthlib``.

		Returns:
			bool: ``True`` if successfully authenticated, ``False`` if not.
		"""

		if not self.token:
			try:
				token = self._token_handler.load()
				token['expires_at'] = time.time() - 10
				self.token = token
			except FileNotFoundError:
				authorization_url = self._session.authorization_url()

				code = input(
					f"Visit:\n\n{authorization_url}\n\n"
					"Follow the prompts and paste provided code: "
				)

				self._session.fetch_token(code)

		self._session.refresh_token()
		self._token_handler.dump(self.token)

		return self.is_authenticated

	# TODO: Revoke oauth token/delete oauth token file.
	def logout(self):
		"""Log out of Google Music."""

		self._session.close()
		self._session = None
		self._username = None
		self._token_handler = None

		try:
			delattr('_uploader_id')
			delattr('_uploader_name')
		except AttributeError:
			pass

		return True

	# TODO Revoke oauth token/delete oauth token file.
	def switch_user(self, username='', *, token=None):
		"""Log in to Google Music with a different user.

		Parameters:
			username (str, Optional):
				Your Google Music username.
				Used to store OAuth tokens for multiple accounts separately.
			token (dict, Optional):
				An OAuth token compatible with ``oauthlib``.

		Returns:
			bool: ``True`` if successfully authenticated, ``False`` if not.
		"""

		if self.logout():
			return self.login(username, token=token)

		return False



__all__ = [
	'FileTokenHandler',
	'TokenHandler',
]

import abc
import json
from pathlib import Path

import appdirs

from .__about__ import __author__, __title__

TOKEN_DIR = Path(appdirs.user_data_dir(__title__, __author__))


class TokenHandler(abc.ABC):
	def __init__(self, **kwargs):
		for k, v in kwargs.items():
			setattr(self, k, v)

	@abc.abstractmethod
	def dump(self, token):
		"""Dump an OAuth token to storage."""

	@abc.abstractmethod
	def load(self):
		"""Load an OAuth token from storage."""


class FileTokenHandler(TokenHandler):
	def dump(self, token, *, username=None, client=None):
		username = username or getattr(self, 'username', '')
		client = client or getattr(self, 'client', '')

		token_path = TOKEN_DIR / username / f'{client}.token'

		try:
			token_path.parent.mkdir(parents=True)
		except FileExistsError:
			pass

		with token_path.open('w') as f:
			json.dump(token, f)

		self.token = token
		self.token_path = token_path

	def load(self, username=None, client=None):
		username = username or getattr(self, 'username', '')
		client = client or getattr(self, 'client', '')

		token_path = TOKEN_DIR / username / f'{client}.token'

		try:
			token_path.parent.mkdir(parents=True)
		except FileExistsError:
			pass

		with token_path.open('r') as f:
			try:
				token = json.load(f)
			except json.JSONDecodeError:
				token = {}

		self.token = token
		self.token_path = token_path

		return self.token



__all__ = [
	'mobileclient',
	'musicmanager',
]

from .clients import MobileClient, MusicManager
from .token_handlers import FileTokenHandler


def mobileclient(
	username=None,
	device_id=None,
	*,
	locale='en_US',
	session=None,
	token=None,
	token_handler=FileTokenHandler,
	token_handler_kwargs=None
):
	"""Create and authenticate a Google Music mobile client.

	>>> import google_music
	>>> mc = google_music.mobileclient('username')

	Parameters:
		username (str, Optional):
			Your Google Music username.
			Used to store OAuth tokens for multiple accounts separately.
		device_id (str, Optional):
			A mobile device ID.
			Default: MAC address is used.
		locale (str, Optional):
			`ICU <http://www.localeplanet.com/icu/>`__
			locale used to localize some responses.
			This must be a locale supported by Android.
			Default: ``'en_US'``.
		session (:class:`~google_music.GoogleMusicSession`, Optional):
			A session compatible with :class:`GoogleMusicSession`.
		token (dict, Optional):
			An OAuth token compatible with ``oauthlib``.
		token_handler (:class:`~google_music.TokenHandler`, Optional):
			A token handler class compatible with :class:`TokenHandler`
			for dumping and loading the OAuth token.
		token_handler_kwargs (dict, Optional):
			Keyword arguments to pass to the ``token_handler``
			class. These become attributes on the class instance.

	Returns:
		MobileClient: An authenticated :class:`~google_music.MobileClient` instance.
	"""

	return MobileClient(
		username,
		device_id,
		locale=locale,
		session=session,
		token=token,
		token_handler=FileTokenHandler,
		token_handler_kwargs=None
	)


def musicmanager(
	username=None,
	uploader_id=None,
	*,
	session=None,
	token=None,
	token_handler=FileTokenHandler,
	token_handler_kwargs=None
):
	"""Create and authenticate a Google Music Music Manager client.

	>>> import google_music
	>>> mm = google_music.musicmanager('username')

	Parameters:
		username (str, Optional):
			Your Google Music username.
			Used to store OAuth tokens for multiple accounts separately.
		uploader_id (str, Optional):
			A unique uploader ID.
			Default: MAC address and username used.
		session (:class:`~google_music.GoogleMusicSession`, Optional):
			A session compatible with :class:`GoogleMusicSession`.
		token (dict, Optional):
			An OAuth token compatible with ``oauthlib``.
		token_handler (:class:`~google_music.TokenHandler`, Optional):
			A token handler class compatible with :class:`TokenHandler`
			for dumping and loading the OAuth token.
		token_handler_kwargs (dict, Optional):
			Keyword arguments to pass to the ``token_handler``
			class. These become attributes on the class instance.

	Returns:
		MusicManager: An authenticated :class:`~google_music.MusicManager` instance.
	"""

	return MusicManager(
		username,
		uploader_id,
		session=session,
		token=token,
		token_handler=FileTokenHandler,
		token_handler_kwargs=None
	)
